// SPDX-License-Identifier: GPL-3.0-or-later

// utility header to include all the message wrappers at once

#ifndef SCHEMA_WRAPPERS_H
#define SCHEMA_WRAPPERS_H

#include "connection.h"
#include "node_connection.h"
#include "node_creation.h"
#include "alarm_config.h"
#include "alarm_stream.h"
#include "node_info.h"
#include "capability.h"
#include "context_stream.h"
#include "context.h"
#include "agent_cmds.h"

#endif /* SCHEMA_WRAPPERS_H */
