// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef TEST_CGROUPS_PLUGIN_H
#define TEST_CGROUPS_PLUGIN_H 1

#include "libnetdata/libnetdata.h"

#include "../sys_fs_cgroup.h"

#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <stdint.h>
#include <cmocka.h>

#endif /* TEST_CGROUPS_PLUGIN_H */
