// DO NOT EDIT THIS FILE DIRECTLY
// It gets generated by integrations/gen_integrations.py in the Netdata repo

export const categories = [[ categories ]]
export const integrations = [[ integrations ]]

