!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},d=(new Error).stack;d&&(e._sentryDebugIds=e._sentryDebugIds||{},e._sentryDebugIds[d]="cd8bda79-8a84-4361-9200-a6dd6284d359",e._sentryDebugIdIdentifier="sentry-dbid-cd8bda79-8a84-4361-9200-a6dd6284d359")}catch(e){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[5969],{65969:function(e,d,n){n.r(d);var a=n(29439),t=n(67294),l=n(61152),o=n(15794),f=n(18663);d.default=function(){var e=(0,l.Z)("manageInvitations"),d=(0,a.Z)(e,4),n=d[0],r=d[3];return t.createElement(o.Z,null,n&&t.createElement(f.l,{onClose:r}))}}}]);