!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},n=(new Error).stack;n&&(e._sentryDebugIds=e._sentryDebugIds||{},e._sentryDebugIds[n]="f2f64cf2-d3e6-4dd3-be20-2371ddaa0e10",e._sentryDebugIdIdentifier="sentry-dbid-f2f64cf2-d3e6-4dd3-be20-2371ddaa0e10")}catch(e){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[8977],{98977:function(e,n,d){d.r(n);d(41539),d(88674),d(17727);var f=d(67294),t=d(1043),u=d(78266),l=d(87267);n.default=function(){var e=(0,l.Z)(),n=e.sendLog,d=e.isReady,o=(0,t.T4)(),a=(0,f.useCallback)((function(){n({feature:"SignUpThankYou"}).finally((function(){o()}))}),[o,d]);return(0,f.useEffect)((function(){var e=!0,n=null;return d?a():n=setTimeout((function(){e&&a()}),1e3),function(){e=!1,n&&(clearTimeout(n),n=null)}}),[d]),f.createElement(u.Z,null)}}}]);