!function(){try{var o="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},t=(new Error).stack;t&&(o._sentryDebugIds=o._sentryDebugIds||{},o._sentryDebugIds[t]="4106ae24-1251-48ca-8980-a45702c556de",o._sentryDebugIdIdentifier="sentry-dbid-4106ae24-1251-48ca-8980-a45702c556de")}catch(o){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[6723],{76723:function(o,t,n){n.r(t),n.d(t,{Notification:function(){return f.P_},Warning:function(){return g},createErrorNotification:function(){return f.t_},createNotification:function(){return f.sc},default:function(){return p},showDashboardCreatedNotification:function(){return u.FR},showInvalidDashboardSlugNotification:function(){return u.cH},showJoiningSpaceNotification:function(){return u.P4},showNodesNotification:function(){return u.Q5},showRoomCreationNotification:function(){return u.R6},showRoomsInSpaceNotification:function(){return u.aN},showRoomsNotification:function(){return u.$T},showSpaceNotification:function(){return u.iL},showUsersInRoomNotification:function(){return u.oG},showUsersInSpaceNotification:function(){return u.OU},toast:function(){return d.Z}});var e=n(87462),r=n(45987),i=n(67294),a=n(71893),c=n(59978),s=n(55678),d=n(52631),u=n(33427),f=n(74855),l=["title","children"],g=function(o){var t=o.title,n=o.children,a=(0,r.Z)(o,l);return i.createElement(c.Flex,(0,e.Z)({column:!0,gap:2,background:"warningBackground",backgroundOpacity:.7,round:!0,padding:[4]},a),t&&i.createElement(c.H3,null,t),n)},b=["className"],p=(0,a.default)((function(o){var t=o.className,n=(0,r.Z)(o,b);return i.createElement(c.Box,{className:t},i.createElement(s.Ix,(0,e.Z)({},n,{closeButton:!1})))})).withConfig({displayName:"notifications__Container",componentId:"sc-8t8ne-0"})([".Toastify__toast-container{position:fixed;width:unset;min-width:400px;max-width:500px;z-index:50;color:",";}.Toastify__toast{padding:0;padding-top:5px;background:",";border:1px solid ",";}.Toastify__toast--error{background:",";border:1px solid ",";}.Toastify__toast--warning{background:",";border:1px solid ",";}.Toastify__toast--success{background:",";border:1px solid ",";}.Toastify__toast-icon{display:none;}.Toastify__toast-body{}.Toastify__progress-bar{bottom:unset;top:0;}.Toastify__progress-bar--success{background-color:",";}.Toastify__progress-bar--error{background-color:",";}"],(0,c.getColor)(["neutral","limedSpruce"]),(0,c.getColor)("elementBackground"),(0,c.getColor)("text"),(0,c.getColor)(["red","lavender"]),(0,c.getColor)("error"),(0,c.getColor)("bright"),(0,c.getColor)("warning"),(0,c.getColor)(["green","frostee"]),(0,c.getColor)("success"),(0,c.getColor)("success"),(0,c.getColor)("error"))}}]);