!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},t=(new Error).stack;t&&(e._sentryDebugIds=e._sentryDebugIds||{},e._sentryDebugIds[t]="9c15662a-734c-4bd8-ac3e-7131c4ae493b",e._sentryDebugIdIdentifier="sentry-dbid-9c15662a-734c-4bd8-ac3e-7131c4ae493b")}catch(e){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[1189],{8937:function(e,t,n){n.r(t);var o=n(87462),d=n(45987),a=n(67294),i=n(5575),r=n(59978),l=n(71893),c=["autoFocus"],f=(0,l.default)(i.default).attrs({"data-testid":"editor"}).withConfig({displayName:"editor__StyledEditor",componentId:"sc-c7iwfz-0"})(["position:relative;outline:none;word-wrap:break-word;&.medium-editor-placeholder:after{color:",";content:attr(data-placeholder) !important;font-style:italic;left:0;position:absolute;top:0;white-space:pre;padding:inherit;margin:inherit;}"],(0,r.getColor)("border"));t.default=function(e){var t=e.autoFocus,n=(0,d.Z)(e,c),i=(0,a.useRef)();return(0,a.useLayoutEffect)((function(){if(t){var e=i.current.medium.elements[0].lastChild;e.scrollIntoView(),i.current.medium.selectElement(e),document.getSelection().collapseToEnd()}}),[]),a.createElement(f,(0,o.Z)({ref:i},n))}}}]);