!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},n=(new Error).stack;n&&(e._sentryDebugIds=e._sentryDebugIds||{},e._sentryDebugIds[n]="5de32fd1-2c07-46d2-9c41-2a966c9dda76",e._sentryDebugIdIdentifier="sentry-dbid-5de32fd1-2c07-46d2-9c41-2a966c9dda76")}catch(e){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[7241],{1178:function(e,n,t){t.r(n),t.d(n,{MagicLinkSent:function(){return p}});t(74916),t(64765);var i=t(67294),a=t(71893),r=t(89250),c=t(59978),o=t(15794),d=t(27266),l=t(87562),u=t(1043),f=t(39979),s=t(63346),m=(0,a.default)(c.Text).attrs({role:"button"}).withConfig({displayName:"magicLinkSent__ButtonText",componentId:"sc-ua6kmo-0"})(["cursor:pointer;"]),g=(0,f.Z)(m),p=function(){var e=(0,r.TH)(),n=e.search,t=e.state,a=((void 0===t?{}:t)||{}).email,o=window.location.hash,f=(0,u.iM)(),m=(0,i.useCallback)((function(){var e=encodeURIComponent((0,l.$)("/sign-in".concat(n),o)),t=encodeURIComponent((0,l.$)("/sign-up/verify".concat(n),o));f({email:a,redirectURI:e,registerURI:t,resend:!0})}),[a]);return i.createElement(s.ZP,{feature:"MagicLinkSent",email:a},i.createElement(d.Z,{"data-testid":"magicLinkSent"},i.createElement(c.H1,{textAlign:"center"},"Check your email!"),i.createElement(c.Flex,{column:!0,padding:[0,0,8,0],gap:8,justifyContent:"between",alignItems:"center"},i.createElement(c.Flex,{column:!0,gap:1,alignItems:"center"},i.createElement(c.TextBig,{textAlign:"center"},"We have sent an email to ",i.createElement(c.TextBig,{strong:!0},!!a&&a),"."),i.createElement(c.TextBig,{textAlign:"center"},"Please find this email (check your spam folder too) and click the button there to continue.")),i.createElement(c.Text,{textAlign:"center"},"Didn't receive it?"," ",i.createElement(g,{onClick:m,color:"primary","data-ga":"magicLikSent::click-resent::check-email-view"},"Click here to resend it.")))))};n.default=(0,o.k)(p,"light")},76362:function(e,n,t){t.d(n,{S1:function(){return r},ZT:function(){return i},ae:function(){return o},jU:function(){return c},on:function(){return a}});var i=function(){};function a(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];e&&e.addEventListener&&e.addEventListener.apply(e,n)}function r(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];e&&e.removeEventListener&&e.removeEventListener.apply(e,n)}var c="undefined"!==typeof window,o="undefined"!==typeof navigator},94829:function(e,n,t){var i=t(67294),a=t(76362);n.Z=function(e,n){var t=(0,i.useState)(function(e,n){return void 0!==n?n:!!a.jU&&window.matchMedia(e).matches}(e,n)),r=t[0],c=t[1];return(0,i.useEffect)((function(){var n=!0,t=window.matchMedia(e),i=function(){n&&c(!!t.matches)};return t.addListener(i),c(t.matches),function(){n=!1,t.removeListener(i)}}),[e]),r}}}]);