!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},n=(new Error).stack;n&&(e._sentryDebugIds=e._sentryDebugIds||{},e._sentryDebugIds[n]="23f9fc85-d4ce-45f3-9769-51f7efea5d5e",e._sentryDebugIdIdentifier="sentry-dbid-23f9fc85-d4ce-45f3-9769-51f7efea5d5e")}catch(e){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[6817],{76817:function(e,n,d){d.r(n),d.d(n,{AuthTokenCallback:function(){return o}});var f=d(67294),t=d(58591),o=function(){var e=(0,t.np)("redirect_uri"),n=(0,t.np)("token");return(0,f.useEffect)((function(){e&&n&&(localStorage.setItem("netdataJWT",n),location.href=decodeURIComponent(e))}),[]),null};n.default=o}}]);