!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},n=(new Error).stack;n&&(e._sentryDebugIds=e._sentryDebugIds||{},e._sentryDebugIds[n]="51ca2d2b-bb53-4699-8541-aaa77de69d1e",e._sentryDebugIdIdentifier="sentry-dbid-51ca2d2b-bb53-4699-8541-aaa77de69d1e")}catch(e){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[4744],{34744:function(e,n,t){t.r(n);var a,d=t(29439),l=(t(66992),t(41539),t(88674),t(78783),t(33948),t(74916),t(64765),t(67294)),o=t(2145),r=t(17563),c=t(89250),f=t(13477),u=t(78266),i=t(92501),s=(0,o.Z)((function(){return Promise.all([t.e(4794),t.e(8584),t.e(39)]).then(t.bind(t,44695))}),"AlertWebview");n.default=(a=(0,i.Pf)((function(){var e=window.location.search,n=r.parse(e),t=n.space_id,a=n.room_id,d=n.node_name,o=n.node_id;return(0,f.Iy)("isLoaded")?l.createElement(l.Suspense,{fallback:l.createElement(u.Z,null)},l.createElement(c.Z5,null,l.createElement(c.AW,{path:"alerts/:alertId",element:l.createElement(s,{spaceId:t,roomId:a,nodeName:d,nodeId:o})}))):l.createElement(u.Z,null)})),function(e){var n=window.location.hash,t=r.parse(n).token,o=(0,l.useState)(),c=(0,d.Z)(o,2),f=c[0],i=c[1];return(0,l.useEffect)((function(){t&&(i(t),localStorage.setItem("netdataJWT",t))}),[]),f?l.createElement(a,e):l.createElement(u.Z,null)})}}]);