!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},n=(new Error).stack;n&&(e._sentryDebugIds=e._sentryDebugIds||{},e._sentryDebugIds[n]="e74f6993-8129-4a9b-9074-e8e2abe5e2a4",e._sentryDebugIdIdentifier="sentry-dbid-e74f6993-8129-4a9b-9074-e8e2abe5e2a4")}catch(e){}}();var _global="undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};_global.SENTRY_RELEASE={id:"0c0a7a41a0e7ecb4870dbce38dbd7d5e0913bb7f"},(self.webpackChunkcloud_frontend=self.webpackChunkcloud_frontend||[]).push([[8447],{38626:function(e,n,t){var o=t(45987),r=t(67294),i=t(59978),f=t(64637),u=t(82351),a=["maxLength","text","TextComponent"];n.Z=function(e){var n=e.maxLength,t=e.text,d=e.TextComponent,c=void 0===d?i.Text:d,s=(0,o.Z)(e,a),l=(0,r.useMemo)((function(){return(0,f.a)(t,n)}),[n,t]);return l&&l.length!==t.length?r.createElement(u.Z,{content:t,align:"bottom",isBasic:!0},r.createElement(c,s,l)):r.createElement(c,s,t)}},76362:function(e,n,t){t.d(n,{S1:function(){return i},ZT:function(){return o},ae:function(){return u},jU:function(){return f},on:function(){return r}});var o=function(){};function r(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];e&&e.addEventListener&&e.addEventListener.apply(e,n)}function i(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];e&&e.removeEventListener&&e.removeEventListener.apply(e,n)}var f="undefined"!==typeof window,u="undefined"!==typeof navigator},11042:function(e,n,t){var o=t(67294),r=t(76362).jU?o.useLayoutEffect:o.useEffect;n.Z=r},88553:function(e,n,t){var o=t(67294),r=t(11042),i=t(76362),f={x:0,y:0,width:0,height:0,top:0,left:0,bottom:0,right:0};n.Z=i.jU&&"undefined"!==typeof window.ResizeObserver?function(){var e=(0,o.useState)(null),n=e[0],t=e[1],i=(0,o.useState)(f),u=i[0],a=i[1],d=(0,o.useMemo)((function(){return new window.ResizeObserver((function(e){if(e[0]){var n=e[0].contentRect,t=n.x,o=n.y,r=n.width,i=n.height,f=n.top,u=n.left,d=n.bottom,c=n.right;a({x:t,y:o,width:r,height:i,top:f,left:u,bottom:d,right:c})}}))}),[]);return(0,r.Z)((function(){if(n)return d.observe(n),function(){d.disconnect()}}),[n]),[t,u]}:function(){return[i.ZT,f]}},94829:function(e,n,t){var o=t(67294),r=t(76362);n.Z=function(e,n){var t=(0,o.useState)(function(e,n){return void 0!==n?n:!!r.jU&&window.matchMedia(e).matches}(e,n)),i=t[0],f=t[1];return(0,o.useEffect)((function(){var n=!0,t=window.matchMedia(e),o=function(){n&&f(!!t.matches)};return t.addListener(o),f(t.matches),function(){n=!1,t.removeListener(o)}}),[e]),i}}}]);